"""
EXSLT 2.0 - Dyanmic (http://www.exslt.org/dyn/index.html)
WWW: http://4suite.org/XSLT        e-mail: support@4suite.org

Copyright (c) 2001 Fourthought Inc, USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

from Ft.Xml.XPath import parser
from Ft.Xml.XPath import Conversions
from Ft.Xml.Xslt import XsltRuntimeException, Error
from Ft.Xml.Xslt.CopyOfElement import CopyNode

EXSL_DYNAMIC_NS = "http://exslt.org/dynamic"

##def Closure(context, nodeset, string):
##    """
##    The dyn:closure function creates a node set resulting from transitive
##    closure of evaluating the expression passed as the second argument on
##    each of the nodes passed as the first argument, then on the node set
##    resulting from that and so on until no more nodes are found.
##
##    Implements version 1.
##    """
##    if type(nodeset) != type([]):
##        raise XsltRuntimeException(Error.WRONG_ARGUMENT_TYPE,
##                                   context.currentInstruction)
##    string = Conversions.StringValue(string)
##    return


def Evaluate(context, string):
    """
    The dyn:evaluate function evaluates a string as an XPath expression and
    returns the resulting value, which might be a boolean, number, string,
    node set, result tree fragment or external object. The sole argument is
    the string to be evaluated.

    Implements version 1.
    """
    string = Conversions.StringValue(string)
    p = parser.new()
    try:
        result = p.parse(string).evaluate(context)
    except:
        import traceback
        traceback.print_exc()
        result = []
    return result


def Map(context, nodeset, string):
    """
    The dyn:map function evaluates the expression passed as the second
    argument for each of the nodes passed as the first argument, and returns
    a node set of those values.

    Implements version 1.
    """
    if type(nodeset) != type([]):
        raise XsltRuntimeException(Error.WRONG_ARGUMENT_TYPE,
                                   context.currentInstruction)
    string = Conversions.StringValue(string)
    p = parser.new()
    ctx = context.clone()
    try:
        result = []
        for n in nodeset:
            ctx.node = n
            result.append(p.parse(string).evaluate(ctx))
    except:
        import traceback
        traceback.print_exc()
        result = []
    return result


##def Max(context, nodeset, string):
##    """
##    The dyn:max function calculates the maximum value for the nodes passed as
##    the first argument, where the value of each node is calculated dynamically
##    using an XPath expression passed as a string as the second argument.
##
##    Implements version 1.
##    """
##    if type(nodeset) != type([]):
##        raise XsltRuntimeException(Error.WRONG_ARGUMENT_TYPE,
##                                   context.currentInstruction)
##    string = Conversions.StringValue(string)
##    return


##def Min(context, nodeset, string):
##    """
##    The dyn:min function calculates the minimum value for the nodes passed as
##    the first argument, where the value of each node is calculated dynamically
##    using an XPath expression passed as a string as the second argument.
##
##    Implements version 1.
##    """
##    if type(nodeset) != type([]):
##        raise XsltRuntimeException(Error.WRONG_ARGUMENT_TYPE,
##                                   context.currentInstruction)
##    string = Conversions.StringValue(string)
##    return


##def Sum(context, nodeset, string):
##    """
##    The dyn:sum function calculates the sum for the nodes passed as the first
##    argument, where the value of each node is calculated dynamically using an
##    XPath expression passed as a string as the second argument.
##
##    Implements version 1.
##    """
##    if type(nodeset) != type([]):
##        raise XsltRuntimeException(Error.WRONG_ARGUMENT_TYPE,
##                                   context.currentInstruction)
##    string = Conversions.StringValue(string)
##    return

ExtNamespaces = {
    EXSL_DYNAMIC_NS : 'dyn',
    }

ExtFunctions = {
    #(EXSL_DYNAMIC_NS, 'closure') : Closure,
    (EXSL_DYNAMIC_NS, 'evaluate') : Evaluate,
    (EXSL_DYNAMIC_NS, 'map') : Map,
    #(EXSL_DYNAMIC_NS, 'max') : Max,
    #(EXSL_DYNAMIC_NS, 'min') : Min,
    #(EXSL_DYNAMIC_NS, 'sum') : Sum,
}

ExtElements = {}



import os, sys
#from Ft.Xml.Xslt import XsltElement, AttributeInfo
from Xml.Xslt import test_harness

XMLDECL = '<?xml version="1.0" encoding="UTF-8"?>\n'

SOURCE_1 = '<doc>&#x2713;</doc>' #Char data is the check mark character

SHEET_1 = """<?xml version="1.0"?>
<xsl:stylesheet version="1.0"
  xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
  xmlns:ext="urn:x-4suite:x"
>

  <xsl:template match="/">
    <xsl:value-of select='ext:ord(.)'/>
  </xsl:template>

</xsl:stylesheet>
"""

EXPECTED_1 = XMLDECL + str(int('2713', 16))

def Test(tester):
    tester.startGroup("Test ord extension function example")

    extmod = '.'.join(__name__.split('.')[:-1] + ['ord'])
    source = test_harness.FileInfo(string=SOURCE_1)
    sty = test_harness.FileInfo(string=SHEET_1)
    test_harness.XsltTest(tester, source, [sty], EXPECTED_1,
                          extensionModules=[extmod])
    tester.groupDone()





import gc
from Ft.Xml import InputSource

def test_empty_node(tester,domMod):

    tester.startTest("Empty Document")
    domMod.StartNodeCounting()
    doc = domMod.implementation.createDocument(None,None,None)
    tester.compare(1,domMod.GetNodeCount())
    del doc
    tester.compare(0,domMod.GetNodeCount())
    tester.testDone()

    tester.startTest("Empty Text")
    domMod.StartNodeCounting()
    doc = domMod.implementation.createDocument(None,None,None)
    text = doc.createTextNode("Foo")
    tester.compare(2,domMod.GetNodeCount())
    del text
    tester.compare(1,domMod.GetNodeCount())
    tester.testDone()

    tester.startTest("Empty Element")
    domMod.StartNodeCounting()
    elem = doc.createElementNS(None,"Foo")
    tester.compare(1,domMod.GetNodeCount())
    del elem
    tester.compare(0,domMod.GetNodeCount())
    tester.testDone()

    tester.startTest("Empty Attribute")
    domMod.StartNodeCounting()
    attr = doc.createAttributeNS(None,"Foo")
    tester.compare(1,domMod.GetNodeCount())
    del attr
    tester.compare(0,domMod.GetNodeCount())
    tester.testDone()

    tester.startTest("Empty Comment")
    domMod.StartNodeCounting()
    com = doc.createComment("Foo")
    tester.compare(1,domMod.GetNodeCount())
    del com
    tester.compare(0,domMod.GetNodeCount())
    tester.testDone()

    tester.startTest("Empty Processing Instruction")
    domMod.StartNodeCounting()
    pi = doc.createProcessingInstruction("Foo","Bar")
    tester.compare(1,domMod.GetNodeCount())
    del pi
    tester.compare(0,domMod.GetNodeCount())
    tester.testDone()

    tester.startTest("Empty Document Fragment")
    domMod.StartNodeCounting()
    df = doc.createDocumentFragment()
    tester.compare(1,domMod.GetNodeCount())
    del df
    tester.compare(0,domMod.GetNodeCount())
    tester.testDone()


def test_small_tree(tester,domMod):
    import sys, gc
    gc.collect()  #Force to clean everything up
    tester.startTest("Single Parent -> child rel")
    doc = domMod.implementation.createDocument(None,None,None)
    domMod.StartNodeCounting()
    elem = doc.createElementNS(None,"Foo")
    elem2 = doc.createElementNS(None,"Foo2")
    elem.appendChild(elem2)
    tester.compare(2,domMod.GetNodeCount())
    del elem
    del elem2
    gc.collect() #Force collection
    tester.compare(0,domMod.GetNodeCount())
    del doc
    tester.testDone()

    tester.startTest("Document -> elem rel")
    domMod.StartNodeCounting()
    doc = domMod.implementation.createDocument(None,None,None)
    elem = doc.createElementNS(None,"Foo")
    doc.appendChild(elem)
    tester.compare(2,domMod.GetNodeCount())
    del doc
    del elem
    gc.collect() #Force collection
    tester.compare(0,domMod.GetNodeCount())
    tester.testDone()

    tester.startTest("Document -> text rel")
    domMod.StartNodeCounting()
    doc = domMod.implementation.createDocument(None,None,None)
    text = doc.createTextNode("Foo")
    doc.appendChild(text)
    tester.compare(2,domMod.GetNodeCount())
    del doc
    del text
    gc.collect() #Force collection
    tester.compare(0,domMod.GetNodeCount())
    tester.testDone()

    tester.startTest("Document -> pi rel")
    domMod.StartNodeCounting()
    doc = domMod.implementation.createDocument(None,None,None)
    pi = doc.createProcessingInstruction("Foo","Bar")
    doc.appendChild(pi)
    tester.compare(2,domMod.GetNodeCount())
    del doc
    del pi
    gc.collect() #Force collection
    tester.compare(0,domMod.GetNodeCount())
    tester.testDone()

    tester.startTest("Document -> comment rel")
    domMod.StartNodeCounting()
    doc = domMod.implementation.createDocument(None,None,None)
    com = doc.createComment("Foo")
    doc.appendChild(com)
    tester.compare(2,domMod.GetNodeCount())
    del doc
    del com
    gc.collect() #Force collection
    tester.compare(0,domMod.GetNodeCount())
    tester.testDone()

def test_df_tree(tester,domMod):
    import sys, gc
    gc.collect()  #Force to clean everything up
    tester.startTest("Document Fragment Tree")
    doc = domMod.implementation.createDocument(None,None,None)
    domMod.StartNodeCounting()
    df = doc.createDocumentFragment()
    elem = doc.createElementNS(None,"Foo")
    elem2 = doc.createElementNS(None,"Foo2")
    df.appendChild(elem)
    df.appendChild(elem2)
    tester.compare(3,domMod.GetNodeCount())
    del elem
    del elem2
    del df
    gc.collect() #Force collection
    tester.compare(0,domMod.GetNodeCount())
    tester.testDone()


def test_attributes(tester,domMod):
    import sys, gc
    gc.collect()  #Force to clean everything up
    tester.startTest("Element with setAttributeNodeNS")
    doc = domMod.implementation.createDocument(None,None,None)
    domMod.StartNodeCounting()
    elem = doc.createElementNS(None,"Foo")
    attr = doc.createAttributeNS(None,"Foo")
    elem.setAttributeNodeNS(attr)
    tester.compare(2,domMod.GetNodeCount())
    del elem
    del attr
    gc.collect() #Force collection
    tester.compare(0,domMod.GetNodeCount())
    tester.testDone()

    tester.startTest("Element with setAttributeNS")
    doc = domMod.implementation.createDocument(None,None,None)
    domMod.StartNodeCounting()
    elem = doc.createElementNS(None,"Foo")
    elem.setAttributeNS(None,"Foo","Bar")
    tester.compare(2,domMod.GetNodeCount())
    del elem
    gc.collect() #Force collection
    #tester.warning("This test fails in cDomlette")
    tester.compare(0,domMod.GetNodeCount())
    tester.testDone()

# -- cyclic garbage collection -----------------------------------------

def test_cycles(tester,domMod):
    tester.startGroup("Reclaiming of Cyclic Memory")
    test_empty_node(tester,domMod)
    test_small_tree(tester,domMod)
    test_df_tree(tester,domMod)
    test_attributes(tester,domMod)
    tester.groupDone()

# -- reference counts --------------------------------------------------

def test_refcounts(tester,domMod):
    tester.startGroup("Low Level Ref Counts")

    tester.startGroup("Internal Created Doc")
    domMod.TestRefCounts(tester,None)
    tester.groupDone()

    tester.startGroup("Empty Doc")
    doc = domMod.implementation.createDocument(None,None,None)
    domMod.TestRefCounts(tester,doc)
    del doc
    gc.collect() #Force collection
    tester.groupDone()

    tester.startGroup("Internal Created Dom")
    domMod.StartNodeCounting()
    dom = domMod.TestTree()
    domMod.TestRefCounts(tester,dom)
    del dom
    gc.collect() #Force collection
    tester.compare(0,domMod.GetNodeCount())
    tester.groupDone()

    tester.startGroup("Hand Created Dom")
    #Becareful to delete all external references exept to the doc
    domMod.StartNodeCounting()
    doc = domMod.implementation.createDocument(None,None,None)
    e = doc.createElementNS('http://foo.com','foo:root')
    doc.appendChild(e)
    
    t = doc.createTextNode("Data1")
    e.appendChild(t)
    del t

    p = doc.createProcessingInstruction("tar","Data3")
    doc.insertBefore(p,e)
    del p

    c = doc.createComment("Data2")
    doc.appendChild(c)
    del c

    e2 = doc.createElementNS('http://foo2.com','foo2:child')
    e.appendChild(e2)
    e2.setAttributeNS('http://foo2.com','foo2:attr','value')
    del e2

    del e
    domMod.TestRefCounts(tester,doc)
    del doc
    gc.collect() #Force collection
    tester.groupDone()

    import test_domlette_readers

    isrc = InputSource.DefaultFactory.fromString(test_domlette_readers.SMALL_XML,'mem')

    tester.startGroup("Small parsed XML")
    domMod.StartNodeCounting()
    dom = domMod.NonvalParse(isrc)
    domMod.TestRefCounts(tester,dom)
    del dom
    gc.collect() #Force collection
    tester.groupDone()

    isrc = InputSource.DefaultFactory.fromString(test_domlette_readers.LARGE_XML,'mem')

    tester.startGroup("Large parsed XML")
    domMod.StartNodeCounting()
    dom = domMod.NonvalParse(isrc)
    domMod.TestRefCounts(tester,dom)
    del dom
    gc.collect() #Force collection
    tester.groupDone()

    isrc = InputSource.DefaultFactory.fromString(test_domlette_readers.SMALL_XML,'mem')

    tester.startGroup("Small parsed XML w/ mod")
    domMod.StartNodeCounting()
    dom = domMod.NonvalParse(isrc)
    for ctr in range(1000):
        dom.documentElement.setAttributeNS('http://foo.com','bar','baz')
    gc.collect() #Force collection

    domMod.TestRefCounts(tester,dom)
    del dom
    gc.collect() #Force collection
    tester.groupDone()

    tester.groupDone()
    return

# -- testing entry function --------------------------------------------

def test_memory(tester, domMod):
    tester.startGroup("Memory")
    test_cycles(tester, domMod)
    #test_refcounts(tester, domMod)
    tester.groupDone()
    return




#ord.py
from Ft.Xml.XPath import Conversions

def Ord(context, s):
    '''
    Available in XPath as ord() as defined by ExtFunctions mapping below
    Takes an object, which is coerced to string
    Returns the Unicode code point number for the first character in that string
    Or returns -1 if it's an empty string
    '''
    s = Conversions.StringValue(s)  #Coerce the passed object to string
    if s:
        return ord(s[0])
    else:
        return -1

ExtFunctions = {
    (u'urn:x-4suite:x', u'ord'): Ord,
}




# demo.py
import time, urlparse
from Ft.Xml.XPath import Conversions

def GetCurrentTime(context):
    '''available in XPath as get-current-time()'''
    return time.asctime(time.localtime())

def HashContextName(context, maxkey):
    '''
    available in XPath as hash-context-name(maxkey),
    where maxkey is an object converted to number
    '''
    # It is a good idea to use the appropriate core function to coerce
    # arguments to the expected type
    maxkey = Conversions.NumberValue(maxkey)
    key = reduce(lambda a, b: a + b, context.node.nodeName)
    return key % maxkey

ExtFunctions = {
    ('urn:x-4suite:x', 'get-current-time'): GetCurrentTime,
    ('urn:x-4suite:x', 'hash-context-name'): HashContextName
}

